/*
 * SMP support for the Xilinx First Stage Boot Loader.
 *
 * Copyright 2013 Chris Johns (chrisj@rtems.org).
 */

#include "fsbl.h"

#if !defined(FSBL_NUM_CORES)
#define FSBL_NUM_CORES 2
#endif

/*
 * SMP enabled ?
 */
static int enabled;

/*
 * SMP support. Get all the cores to come here and wait.
 */
static u8 Secondary_cores[FSBL_NUM_CORES - 1];

/*
 * The type of the entry point.
 */
typedef void (*Fsbl_SMP_entry_point_t)(void);

/*
 * The entry point we branch each core to.
 */
static Fsbl_SMP_entry_point_t Secondary_entry_point;

static inline u32
Fsbl_SMP_Get_current_processor(void)
{
  u32 mpidr;
  /* Use ARMv7 Multiprocessor Affinity Register (MPIDR) */
  __asm__ volatile ("mrc p15, 0, %[mpidr], c0, c0, 5\n" : [mpidr] "=&r" (mpidr));
  return mpidr & 0xffU;
}

static inline void
Fsbl_SMP_Data_sync_barrier(void)
{
  __asm__ volatile ( "dsb" : : : "memory" );
}

static inline void
Fsbl_SMP_Instruction_sync_barrier(void)
{
  __asm__ volatile ( "isb" : : : "memory" );
}

static inline void
Fsbl_SMP_Wait_for_event(void)
{
  __asm__ volatile ( "wfe" : : : "memory" );
}

static inline void
Fsbl_SMP_Send_event(void)
{
  __asm__ volatile ( "sev" : : : "memory" );
}

/**
* The entry point for a secondary core. Flag the core as present.
*/
static void FsblSMPSecondaryCore(void) __attribute__ ((noreturn));
static void
FsblSMPSecondaryCore(void)
{
  u32 cpuid = Fsbl_SMP_Get_current_processor();

  Secondary_cores[cpuid - 1] = 1;

  Fsbl_SMP_Data_sync_barrier();
  Fsbl_SMP_Instruction_sync_barrier();

  while (1)
  {
    if (Secondary_entry_point)
      Secondary_entry_point();
    Fsbl_SMP_Wait_for_event();
  }
}

void
FsblSMPEnable(void)
{
  enabled = 1;
}

void
FsblSMPWakeSecondaryCores(void)
{
  /*
   * The primary processors is the one calling this function so no need for an
   * address to kick it.
   */
  volatile u32* const kick_address[FSBL_NUM_CORES - 1] = {
    (u32*) 0xfffffff0UL
  };
  int cpu;
  int all_working = 0;

  if (!enabled)
    return;

  Fsbl_SMP_Data_sync_barrier();
  Fsbl_SMP_Instruction_sync_barrier();

  for (cpu = 0; cpu < (FSBL_NUM_CORES - 1); ++cpu)
    Secondary_cores[cpu] = 0;

  Fsbl_SMP_Data_sync_barrier();
  Fsbl_SMP_Instruction_sync_barrier();

  for (cpu = 0; cpu < (FSBL_NUM_CORES - 1); ++cpu)
  {
    *(kick_address[cpu]) = (u32) FsblSMPSecondaryCore;
    Fsbl_SMP_Data_sync_barrier();
    Fsbl_SMP_Instruction_sync_barrier();
    Fsbl_SMP_Send_event();
  }

  while (!all_working)
  {
    all_working = 1;
    for (cpu = 0; cpu < (FSBL_NUM_CORES - 1); ++cpu)
    {
      Fsbl_SMP_Data_sync_barrier();
      Fsbl_SMP_Instruction_sync_barrier();

      if (Secondary_cores[cpu] == 0)
        all_working = 0;
    }
  }
}

void
FsblSMPDispatchSecondaryCores(u32 entry_point)
{
  if (!enabled)
    return;
  Secondary_entry_point = (Fsbl_SMP_entry_point_t) entry_point;
  Fsbl_SMP_Send_event();
}
